import os
from typing import Any


__all__ = [
    "fmt_cyan",
    "fmt_green",
    "print_cyan",
    "print_green",
    "print_red",
]


USING_WINDOWS = os.name == "nt"


def print_red(x: Any) -> None:
    if USING_WINDOWS:
        print(x)
    else:
        print(f"\033[91m {x}\033[00m")


def print_green(x: Any) -> None:
    if USING_WINDOWS:
        print(x)
    else:
        print(f"\033[92m {x}\033[00m")


def print_cyan(x: Any) -> None:
    if USING_WINDOWS:
        print(x)
    else:
        print(f"\033[96m {x}\033[00m")


def fmt_green(x: Any) -> str:
    if USING_WINDOWS:
        return x
    else:
        return f"\033[92m {x}\033[00m"


def fmt_cyan(x: Any) -> str:
    if USING_WINDOWS:
        return x
    else:
        return f"\033[96m {x}\033[00m"



"""This module defines the `IoManager` class
which manages I/O for file objects connected to an existing gdb process
or pty.
"""
import logging
import os
import select
import time
from pprint import pformat
from typing import IO, Any, Dict, List, Optional, Tuple, Union

from pygdbmi import gdbmiparser
from pygdbmi.constants import (
    DEFAULT_GDB_TIMEOUT_SEC,
    DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC,
    USING_WINDOWS,
    GdbTimeoutError,
)


if USING_WINDOWS:
    import msvcrt
    from ctypes import POINTER, WinError, byref, windll, wintypes  # type: ignore
    from ctypes.wintypes import BOOL, DWORD, HANDLE
else:
    import fcntl


__all__ = ["IoManager"]


logger = logging.getLogger(__name__)


class IoManager:
    def __init__(
        self,
        stdin: IO[bytes],
        stdout: IO[bytes],
        stderr: Optional[IO[bytes]],
        time_to_check_for_additional_output_sec: float = DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC,
    ) -> None:
        """
        Manage I/O for file objects created before calling this class
        This can be useful if the gdb process is managed elsewhere, or if a
        pty is used.
        """

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

        self.stdin_fileno = self.stdin.fileno()
        self.stdout_fileno = self.stdout.fileno()
        self.stderr_fileno = self.stderr.fileno() if self.stderr else -1

        self.read_list: List[int] = []
        if self.stdout:
            self.read_list.append(self.stdout_fileno)
        self.write_list = [self.stdin_fileno]

        self._incomplete_output: Dict[str, Any] = {"stdout": None, "stderr": None}
        self.time_to_check_for_additional_output_sec = (
            time_to_check_for_additional_output_sec
        )
        self._allow_overwrite_timeout_times = (
            self.time_to_check_for_additional_output_sec > 0
        )
        _make_non_blocking(self.stdout)
        if self.stderr:
            _make_non_blocking(self.stderr)

    def get_gdb_response(
        self,
        timeout_sec: float = DEFAULT_GDB_TIMEOUT_SEC,
        raise_error_on_timeout: bool = True,
    ) -> List[Dict]:
        """Get response from GDB, and block while doing so. If GDB does not have any response ready to be read
        by timeout_sec, an exception is raised.

        Args:
            timeout_sec: Maximum time to wait for reponse. Must be >= 0. Will return after
            raise_error_on_timeout: Whether an exception should be raised if no response was found after timeout_sec

        Returns:
            List of parsed GDB responses, returned from gdbmiparser.parse_response, with the
            additional key 'stream' which is either 'stdout' or 'stderr'

        Raises:
            GdbTimeoutError: if response is not received within timeout_sec
            ValueError: if select returned unexpected file number
        """

        if timeout_sec < 0:
            logger.warning("timeout_sec was negative, replacing with 0")
            timeout_sec = 0

        if USING_WINDOWS:
            retval = self._get_responses_windows(timeout_sec)
        else:
            retval = self._get_responses_unix(timeout_sec)

        if not retval and raise_error_on_timeout:
            raise GdbTimeoutError(
                "Did not get response from gdb after %s seconds" % timeout_sec
            )

        else:
            return retval

    def _get_responses_windows(self, timeout_sec: float) -> List[Dict]:
        """Get responses on windows. Assume no support for select and use a while loop."""
        timeout_time_sec = time.time() + timeout_sec
        responses = []
        while True:
            responses_list = []
            try:
                self.stdout.flush()
                raw_output = self.stdout.readline().replace(b"\r", b"\n")
                responses_list = self._get_responses_list(raw_output, "stdout")
            except OSError:
                pass

            if self.stderr is not None:
                try:
                    self.stderr.flush()
                    raw_output = self.stderr.readline().replace(b"\r", b"\n")
                    responses_list += self._get_responses_list(raw_output, "stderr")
                except OSError:
                    pass

            responses += responses_list
            if timeout_sec == 0:
                break
            elif responses_list and self._allow_overwrite_timeout_times:
                timeout_time_sec = min(
                    time.time() + self.time_to_check_for_additional_output_sec,
                    timeout_time_sec,
                )
            elif time.time() > timeout_time_sec:
                break

        return responses

    def _get_responses_unix(self, timeout_sec: float) -> List[Dict]:
        """Get responses on unix-like system. Use select to wait for output."""
        timeout_time_sec = time.time() + timeout_sec
        responses = []
        while True:
            select_timeout = timeout_time_sec - time.time()
            if select_timeout <= 0:
                select_timeout = 0
            events, _, _ = select.select(self.read_list, [], [], select_timeout)
            responses_list = None  # to avoid infinite loop if using Python 2
            for fileno in events:
                # new data is ready to read
                if fileno == self.stdout_fileno:
                    self.stdout.flush()
                    raw_output = self.stdout.read()
                    stream = "stdout"

                elif fileno == self.stderr_fileno:
                    assert self.stderr is not None
                    self.stderr.flush()
                    raw_output = self.stderr.read()
                    stream = "stderr"

                else:
                    raise ValueError(
                        "Developer error. Got unexpected file number %d" % fileno
                    )
                responses_list = self._get_responses_list(raw_output, stream)
                responses += responses_list

            if timeout_sec == 0:  # just exit immediately
                break

            elif responses_list and self._allow_overwrite_timeout_times:
                # update timeout time to potentially be closer to now to avoid lengthy wait times when nothing is being output by gdb
                timeout_time_sec = min(
                    time.time() + self.time_to_check_for_additional_output_sec,
                    timeout_time_sec,
                )

            elif time.time() > timeout_time_sec:
                break

        return responses

    def _get_responses_list(
        self, raw_output: bytes, stream: str
    ) -> List[Dict[Any, Any]]:
        """Get parsed response list from string output
        Args:
            raw_output (unicode): gdb output to parse
            stream (str): either stdout or stderr
        """
        responses: List[Dict[Any, Any]] = []

        (_new_output, self._incomplete_output[stream],) = _buffer_incomplete_responses(
            raw_output, self._incomplete_output.get(stream)
        )

        if not _new_output:
            return responses

        response_list = list(
            filter(lambda x: x, _new_output.decode(errors="replace").split("\n"))
        )  # remove blank lines

        # parse each response from gdb into a dict, and store in a list
        for response in response_list:
            if gdbmiparser.response_is_finished(response):
                pass
            else:
                parsed_response = gdbmiparser.parse_response(response)
                parsed_response["stream"] = stream

                logger.debug("%s", pformat(parsed_response))

                responses.append(parsed_response)

        return responses

    def write(
        self,
        mi_cmd_to_write: Union[str, List[str]],
        timeout_sec: float = DEFAULT_GDB_TIMEOUT_SEC,
        raise_error_on_timeout: bool = True,
        read_response: bool = True,
    ) -> List[Dict]:
        """Write to gdb process. Block while parsing responses from gdb for a maximum of timeout_sec.

        Args:
            mi_cmd_to_write: String to write to gdb. If list, it is joined by newlines.
            timeout_sec: Maximum number of seconds to wait for response before exiting. Must be >= 0.
            raise_error_on_timeout: If read_response is True, raise error if no response is received
            read_response: Block and read response. If there is a separate thread running, this can be false, and the reading thread read the output.
        Returns:
            List of parsed gdb responses if read_response is True, otherwise []
        Raises:
            TypeError: if mi_cmd_to_write is not valid
        """
        # self.verify_valid_gdb_subprocess()
        if timeout_sec < 0:
            logger.warning("timeout_sec was negative, replacing with 0")
            timeout_sec = 0

        # Ensure proper type of the mi command
        if isinstance(mi_cmd_to_write, str):
            mi_cmd_to_write_str = mi_cmd_to_write
        elif isinstance(mi_cmd_to_write, list):
            mi_cmd_to_write_str = "\n".join(mi_cmd_to_write)
        else:
            raise TypeError(
                "The gdb mi command must a be str or list. Got "
                + str(type(mi_cmd_to_write))
            )

        logger.debug("writing: %s", mi_cmd_to_write)

        if not mi_cmd_to_write_str.endswith("\n"):
            mi_cmd_to_write_nl = mi_cmd_to_write_str + "\n"
        else:
            mi_cmd_to_write_nl = mi_cmd_to_write_str

        if USING_WINDOWS:
            # select not implemented in windows for pipes
            # assume it's always ready
            outputready = [self.stdin_fileno]
        else:
            _, outputready, _ = select.select([], self.write_list, [], timeout_sec)
        for fileno in outputready:
            if fileno == self.stdin_fileno:
                # ready to write
                self.stdin.write(mi_cmd_to_write_nl.encode())  # type: ignore
                # must flush, otherwise gdb won't realize there is data
                # to evaluate, and we won't get a response
                self.stdin.flush()  # type: ignore
            else:
                logger.error("got unexpected fileno %d" % fileno)

        if read_response is True:
            return self.get_gdb_response(
                timeout_sec=timeout_sec, raise_error_on_timeout=raise_error_on_timeout
            )

        else:
            return []


def _buffer_incomplete_responses(
    raw_output: Optional[bytes], buf: Optional[bytes]
) -> Tuple[Optional[bytes], Optional[bytes]]:
    """It is possible for some of gdb's output to be read before it completely finished its response.
    In that case, a partial mi response was read, which cannot be parsed into structured data.
    We want to ALWAYS parse complete mi records. To do this, we store a buffer of gdb's
    output if the output did not end in a newline.

    Args:
        raw_output: Contents of the gdb mi output
        buf (str): Buffered gdb response from the past. This is incomplete and needs to be prepended to
        gdb's next output.

    Returns:
        (raw_output, buf)
    """

    if raw_output:
        if buf:
            # concatenate buffer and new output
            raw_output = b"".join([buf, raw_output])
            buf = None

        if b"\n" not in raw_output:
            # newline was not found, so assume output is incomplete and store in buffer
            buf = raw_output
            raw_output = None

        elif not raw_output.endswith(b"\n"):
            # raw output doesn't end in a newline, so store everything after the last newline (if anything)
            # in the buffer, and parse everything before it
            remainder_offset = raw_output.rindex(b"\n") + 1
            buf = raw_output[remainder_offset:]
            raw_output = raw_output[:remainder_offset]

    return (raw_output, buf)


def _make_non_blocking(file_obj: IO) -> None:
    """make file object non-blocking
    Windows doesn't have the fcntl module, but someone on
    stack overflow supplied this code as an answer, and it works
    http://stackoverflow.com/a/34504971/2893090"""

    if USING_WINDOWS:
        LPDWORD = POINTER(DWORD)
        PIPE_NOWAIT = wintypes.DWORD(0x00000001)

        SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
        SetNamedPipeHandleState.argtypes = [HANDLE, LPDWORD, LPDWORD, LPDWORD]
        SetNamedPipeHandleState.restype = BOOL

        h = msvcrt.get_osfhandle(file_obj.fileno())  # type: ignore

        res = windll.kernel32.SetNamedPipeHandleState(h, byref(PIPE_NOWAIT), None, None)
        if res == 0:
            raise ValueError(WinError())

    else:
        # Set the file status flag (F_SETFL) on the pipes to be non-blocking
        # so we can attempt to read from a pipe with no new data without locking
        # the program up
        fcntl.fcntl(file_obj, fcntl.F_SETFL, os.O_NONBLOCK)



import os


__all__ = [
    "DEFAULT_GDB_TIMEOUT_SEC",
    "DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC",
    "GdbTimeoutError",
    "USING_WINDOWS",
]


DEFAULT_GDB_TIMEOUT_SEC = 1
DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC = 0.2
USING_WINDOWS = os.name == "nt"


class GdbTimeoutError(ValueError):
    """Raised when no response is recieved from gdb after the timeout has been triggered"""

    pass



"""
Support for unescaping strings produced by GDB MI.
"""

import re
from typing import Iterator, Tuple


__all__ = [
    "advance_past_string_with_gdb_escapes",
    "unescape",
]


def unescape(escaped_str: str) -> str:
    """Unescape a string escaped by GDB in MI mode.

    Args:
        escaped_str: String to unescape (without initial and final double quote).

    Returns:
        The strings with escape codes transformed into normal characters.
    """
    unescaped_str, after_string_index = _unescape_internal(
        escaped_str, expect_closing_quote=False
    )
    assert after_string_index == -1, (
        f"after_string_index is {after_string_index} but it was "
        "expected to be -1 as expect_closing_quote is set to False"
    )
    return unescaped_str


def advance_past_string_with_gdb_escapes(
    escaped_str: str, *, start: int = 0
) -> Tuple[str, int]:
    """Unescape a string escaped by GDB in MI mode, and find the double quote
    terminating it.

    Args:
        escaped_str: String to unescape (without initial double quote).
        start: the position in escaped_str at which to start unescaping the string

    Returns:
        A tuple containing the unescaped string and the index in escaped_str just after
        the escape string (that is, escaped_str[start-1] is always the closing double
        quote and escaped_str[start:] is the portion of escaped_str after the escaped
        string).
    """
    return _unescape_internal(escaped_str, expect_closing_quote=True, start=start)


# Regular expression matching both escapes and unescaped quotes in GDB MI escaped
# strings.
_ESCAPES_RE = re.compile(
    r"""
    # Match all text before an escape or quote so it can be preserved as is.
    (?P<before>
        .*?
    )
    # Match either an escape or an unescaped quote.
    (
        (
            # All escapes start with a backslash...
            \\
            # ... and are followed by either a 3-digit octal number or a single
            # character for common escapes. See _GDB_MI_NON_OCTAL_ESCAPES for valid
            # ones.
            (
                # Instead of matching a single octal escape we match multiple ones in a
                # row.
                # This is because a single Unicode character can be encoded as multiple
                # escape sequences so, if we decoded the escape sequences one at a time,
                # the resulting string would not be valid until all the bytes are
                # converted.
                # This could also be solved by converting the input string into bytes
                # but that's much slower for long strings.
                (?P<escaped_octal>
                    # First octal number without backslash which we matched earlier.
                    [0-7]{3}
                    # Addional (and optional) octal numbers, including a backslash.
                    (
                        \\
                        [0-7]{3}
                    )*
                )
                |
                (?P<escaped_char>.)
            )
        )
        |
        # Match an unescaped quote.
        # If expect_closing_quote is true, then this means the string is finished.
        # If false, then the quote should have been escaped.
        (?P<unescaped_quote>")
    )
    """,
    flags=re.VERBOSE,
)

# Map from single character escape codes allowed in GDB MI strings to the corresponding
# unescaped value.
_NON_OCTAL_ESCAPES = {
    "'": "'",
    "\\": "\\",
    "a": "\a",
    "b": "\b",
    "e": "\033",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
    '"': '"',
}


def _unescape_internal(
    escaped_str: str, *, expect_closing_quote: bool, start: int = 0
) -> Tuple[str, int]:
    """Common code for unescaping strings escaped by GDB in MI mode.

    MI-mode escapes are similar to standard Python escapes but:
    * "\\e" is a valid escape.
    * "\\NNN" escapes use numbers represented in octal format.
      For instance, "\\040" encodes character 0o40, that is character 32 in decimal,
      that is a space.

    For details, see printchar in gdb/utils.c in the binutils-gdb repo.

    Args:
        escaped_str: String to unescape
        expect_closing_quote: If true the closing quote must be in escaped_str[start:].
            Otherwise, no unescaped quote is allowed.
        start: the position in escaped_str at which to start unescaping the string.

    Returns:
        A tuple containing the unescaped string and the index in escaped_str just after
        the escape string, or -1 if expect_closing_quote is False.
    """
    # The _ESCAPES_RE expression only matches escapes or unescaped quotes, plus the
    # preeeding part of the escaped string.
    # This variable tracks the end of the last match so the portion of escaped_str after
    # that is not lost.
    unmatched_start_index = start

    # Was the closing quote found?
    # This can be true only if expect_closing_quote is true.
    found_closing_quote = False

    unescaped_parts = []
    for match in _ESCAPES_RE.finditer(escaped_str, pos=start):
        # Text before the match (and after any previous match).
        unescaped_parts.append(match["before"])

        escaped_octal = match["escaped_octal"]
        escaped_char = match["escaped_char"]
        unescaped_quote = match["unescaped_quote"]

        _, unmatched_start_index = match.span()

        if escaped_octal is not None:
            # We found one or more octal escapes. These are in the form "NNN" or, for
            # multiple characters in a row, "NNN\NNN\NNN[...]".
            # escaped_octal is guaranteed to be in the correct format by _ESCAPES_RE.
            octal_sequence_bytes = bytearray()
            # Strip the backslashes and iterate over the octal codes 3 by 3.
            for octal_number in _split_n_chars(escaped_octal.replace("\\", ""), 3):
                # Convert the 3 digits into a single byte.
                try:
                    octal_sequence_bytes.append(int(octal_number, base=8))
                except ValueError as exc:
                    raise ValueError(
                        f"Invalid octal number {octal_number!r} in {escaped_str!r}"
                    ) from exc
            try:
                replaced = octal_sequence_bytes.decode("utf-8")
            except UnicodeDecodeError:
                # GDB should never generate invalid sequences but, according to #64,
                # it can do that on Windows. In this case we just keep the sequence
                # unchanged.
                replaced = f"\\{escaped_octal}"

        elif escaped_char is not None:
            # We found a single escaped character.
            try:
                replaced = _NON_OCTAL_ESCAPES[escaped_char]
            except KeyError as exc:
                raise ValueError(
                    f"Invalid escape character {escaped_char!r} in {escaped_str!r}"
                ) from exc

        elif unescaped_quote:
            # We found an unescaped quote.
            if not expect_closing_quote:
                raise ValueError(f"Unescaped quote found in {escaped_str!r}")

            # This is the ending quote, so stop processing.
            found_closing_quote = True
            break

        else:
            raise AssertionError(
                f"This code should not be reached for string {escaped_str!r}"
            )

        unescaped_parts.append(replaced)

    if not found_closing_quote:
        if expect_closing_quote:
            raise ValueError(f"Missing closing quote in {escaped_str!r}")

        # Don't drop the part of the escaped string after the last escape.
        unescaped_parts.append(escaped_str[unmatched_start_index:])
        # With expect_closing_quote being false, the whole string must always be matched
        # so unmatched_start_index is not useful so we set it to -1.
        # (We could set it to len(unmatched_start_index) as well but we would not get
        # any benefit from having it set to a correct value.)
        unmatched_start_index = -1

    return "".join(unescaped_parts), unmatched_start_index


def _split_n_chars(s: str, n: int) -> Iterator[str]:
    """Iterates over string s `n` characters at a time"""
    for i in range(0, len(s), n):
        yield s[i : i + n]



"""
Python parser for gdb's machine interface interpreter.

Parses string output from gdb with the `--interpreter=mi2` flag into
structured objects.

See more at https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html#GDB_002fMI
"""

import functools
import logging
import re
from typing import Any, Callable, Dict, List, Match, Optional, Pattern, Tuple, Union

from pygdbmi.gdbescapes import unescape
from pygdbmi.printcolor import fmt_green
from pygdbmi.StringStream import StringStream


__all__ = [
    "parse_response",
    "response_is_finished",
]


_DEBUG = False
logger = logging.getLogger(__name__)


def _setup_logger(logger: logging.Logger, debug: bool) -> None:
    logger.propagate = False

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("[%(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s")
    )
    if debug:
        level = logging.DEBUG
    else:
        level = logging.ERROR

    logger.setLevel(level)
    logger.addHandler(handler)


_setup_logger(logger, _DEBUG)


def parse_response(gdb_mi_text: str) -> Dict:
    """Parse gdb mi text and turn it into a dictionary.

    See https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
    for details on types of gdb mi output.

    Args:
        gdb_mi_text: String output from gdb

    Returns:
        dictionary with keys "type", "message", "payload", "token"
    """
    stream = StringStream(gdb_mi_text, debug=_DEBUG)

    for pattern, parser in _GDB_MI_PATTERNS_AND_PARSERS:
        match = pattern.match(gdb_mi_text)
        if match is not None:
            return parser(match, stream)

    # This was not gdb mi output, so it must have just been printed by
    # the inferior program that's being debugged
    return {
        "type": "output",
        "message": None,
        "payload": gdb_mi_text,
    }


def response_is_finished(gdb_mi_text: str) -> bool:
    """Return true if the gdb mi response is ending

    Args:
        gdb_mi_text: String output from gdb

    Returns:
        True if gdb response is finished
    """
    return _GDB_MI_RESPONSE_FINISHED_RE.match(gdb_mi_text) is not None


# ========================================================================
# All functions and variables below are used internally to parse mi output
# ========================================================================


def _parse_mi_notify(match: Match, stream: StringStream) -> Dict:
    """Parser function for matches against a notify record.

    See _GDB_MI_PATTERNS_AND_PARSERS for details."""
    message = match["message"]
    logger.debug("parsed message")
    logger.debug("%s", fmt_green(message))

    return {
        "type": "notify",
        "message": message.strip(),
        "payload": _extract_payload(match, stream),
        "token": _extract_token(match),
    }


def _parse_mi_result(match: Match, stream: StringStream) -> Dict:
    """Parser function for matches against a result record.

    See _GDB_MI_PATTERNS_AND_PARSERS for details."""
    return {
        "type": "result",
        "message": match["message"],
        "payload": _extract_payload(match, stream),
        "token": _extract_token(match),
    }


def _parse_mi_output(match: Match, stream: StringStream, output_type: str) -> Dict:
    """Parser function for matches against a console, log or target record.

    The record type must be specified in output_type.

    See _GDB_MI_PATTERNS_AND_PARSERS for details."""
    return {
        "type": output_type,
        "message": None,
        "payload": unescape(match["payload"]),
    }


def _parse_mi_finished(match: Match, stream: StringStream) -> Dict:
    """Parser function for matches against a finished record.

    See _GDB_MI_PATTERNS_AND_PARSERS for details."""
    return {
        "type": "done",
        "message": None,
        "payload": None,
    }


def _extract_token(match: Match) -> Optional[int]:
    """Extract a token from a match against a regular expression which included
    _GDB_MI_COMPONENT_TOKEN."""
    token = match["token"]
    return int(token) if token is not None else None


def _extract_payload(match: Match, stream: StringStream) -> Optional[Dict]:
    """Extract a token from a match against a regular expression which included
    _GDB_MI_COMPONENT_PAYLOAD."""
    if match["payload"] is None:
        return None

    stream.advance_past_chars([","])
    return _parse_dict(stream)


# A regular expression matching a response finished record.
_GDB_MI_RESPONSE_FINISHED_RE = re.compile(r"^\(gdb\)\s*$")

# Regular expression identifying a token in a MI record.
_GDB_MI_COMPONENT_TOKEN = r"(?P<token>\d+)?"
# Regular expression identifying a payload in a MI record.
_GDB_MI_COMPONENT_PAYLOAD = r"(?P<payload>,.*)?"

# The type of the functions which parse MI records as used by
# _GDB_MI_PATTERNS_AND_PARSERS.
_PARSER_FUNCTION = Callable[[Match, StringStream], Dict]

# A list where each item is a tuple of:
# - A compiled regular expression matching a MI record.
# - A function which is called if the regex matched with the match and a StringStream.
#   It must return a dictionary with details on the MI record..
#
# For more details on the MI , see
# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
#
# The order matters as items are iterated in ordered and that stops once a match is
# found.
_GDB_MI_PATTERNS_AND_PARSERS: List[Tuple[Pattern, _PARSER_FUNCTION]] = [
    # https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Result-Records.html#GDB_002fMI-Result-Records
    # In addition to a number of out-of-band notifications,
    # the response to a gdb/mi command includes one of the following result indications:
    # done, running, connected, error, exit
    (
        re.compile(
            rf"^{_GDB_MI_COMPONENT_TOKEN}\^(?P<message>\S+?){_GDB_MI_COMPONENT_PAYLOAD}$"
        ),
        _parse_mi_result,
    ),
    # https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Async-Records.html#GDB_002fMI-Async-Records
    # Async records are used to notify the gdb/mi client of additional
    # changes that have occurred. Those changes can either be a consequence
    # of gdb/mi commands (e.g., a breakpoint modified) or a result of target activity
    # (e.g., target stopped).
    (
        re.compile(
            rf"^{_GDB_MI_COMPONENT_TOKEN}[*=](?P<message>\S+?){_GDB_MI_COMPONENT_PAYLOAD}$"
        ),
        _parse_mi_notify,
    ),
    # https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
    # "~" string-output
    # The console output stream contains text that should be displayed
    # in the CLI console window. It contains the textual responses to CLI commands.
    (
        re.compile(r'~"(?P<payload>.*)"', re.DOTALL),
        functools.partial(_parse_mi_output, output_type="console"),
    ),
    # https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
    # "&" string-output
    # The log stream contains debugging messages being produced by gdb's internals.
    (
        re.compile(r'&"(?P<payload>.*)"', re.DOTALL),
        functools.partial(_parse_mi_output, output_type="log"),
    ),
    # https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
    # "@" string-output
    # The target output stream contains any textual output from the
    # running target. This is only present when GDB's event loop is truly asynchronous,
    # which is currently only the case for remote targets.
    (
        re.compile(r'@"(?P<payload>.*)"', re.DOTALL),
        functools.partial(_parse_mi_output, output_type="target"),
    ),
    (
        _GDB_MI_RESPONSE_FINISHED_RE,
        _parse_mi_finished,
    ),
]


_WHITESPACE = [" ", "\t", "\r", "\n"]

_GDB_MI_CHAR_DICT_START = "{"
_GDB_MI_CHAR_ARRAY_START = "["
_GDB_MI_CHAR_STRING_START = '"'
_GDB_MI_VALUE_START_CHARS = [
    _GDB_MI_CHAR_DICT_START,
    _GDB_MI_CHAR_ARRAY_START,
    _GDB_MI_CHAR_STRING_START,
]


def _parse_dict(stream: StringStream) -> Dict:
    """Parse dictionary, with optional starting character '{'
    return (tuple):
        Number of characters parsed from to_parse
        Parsed dictionary
    """
    obj: Dict[str, Union[str, list, dict]] = {}

    logger.debug("%s", fmt_green("parsing dict"))

    while True:
        c = stream.read(1)
        if c in _WHITESPACE:
            pass
        elif c in ["{", ","]:
            pass
        elif c in ["}", ""]:
            # end of object, exit loop
            break

        else:
            stream.seek(-1)
            key, val = _parse_key_val(stream)
            if key in obj:
                # This is a gdb bug. We should never get repeated keys in a dict!
                # See https://sourceware.org/bugzilla/show_bug.cgi?id=22217
                # and https://github.com/cs01/pygdbmi/issues/19
                # Example:
                #   thread-ids={thread-id="1",thread-id="2"}
                # Results in:
                #   thread-ids: {{'thread-id': ['1', '2']}}
                # Rather than the lossy
                #   thread-ids: {'thread-id': 2}  # '1' got overwritten!
                if isinstance(obj[key], list):
                    obj[key].append(val)  # type: ignore
                else:
                    obj[key] = [obj[key], val]
            else:
                obj[key] = val

            look_ahead_for_garbage = True
            c = stream.read(1)
            while look_ahead_for_garbage:
                if c in ["}", ",", ""]:
                    look_ahead_for_garbage = False
                else:
                    # got some garbage text, skip it. for example:
                    # name="gdb"gargage  # skip over 'garbage'
                    # name="gdb"\n  # skip over '\n'
                    logger.debug("skipping unexpected charcter: " + c)
                    c = stream.read(1)
            stream.seek(-1)

    logger.debug("parsed dict")
    logger.debug("%s", fmt_green(obj))
    return obj


def _parse_key_val(stream: StringStream) -> Tuple[str, Union[str, List, Dict]]:
    """Parse key, value combination
    return (tuple):
        Parsed key (string)
        Parsed value (either a string, array, or dict)
    """

    logger.debug("parsing key/val")
    key = _parse_key(stream)
    val = _parse_val(stream)

    logger.debug("parsed key/val")
    logger.debug("%s", fmt_green(key))
    logger.debug("%s", fmt_green(val))

    return key, val


def _parse_key(stream: StringStream) -> str:
    """Parse key, value combination
    returns :
        Parsed key (string)
    """
    logger.debug("parsing key")

    key = stream.advance_past_chars(["="])

    logger.debug("parsed key:")
    logger.debug("%s", fmt_green(key))
    return key


def _parse_val(stream: StringStream) -> Union[str, List, Dict]:
    """Parse value from string
    returns:
        Parsed value (either a string, array, or dict)
    """

    logger.debug("parsing value")

    val: Any

    while True:
        c = stream.read(1)

        if c == "{":
            # Start object
            val = _parse_dict(stream)
            break

        elif c == "[":
            # Start of an array
            val = _parse_array(stream)
            break

        elif c == '"':
            # Start of a string
            val = stream.advance_past_string_with_gdb_escapes()
            break

        elif _DEBUG:
            raise ValueError("unexpected character: %s" % c)

        else:
            logger.warn(f'unexpected character: "{c}" ({ord(c)}). Continuing.')
            val = ""  # this will be overwritten if there are more characters to be read

    logger.debug("parsed value:")
    logger.debug("%s", fmt_green(val))

    return val


def _parse_array(stream: StringStream) -> list:
    """Parse an array, stream should be passed the initial [
    returns:
        Parsed array
    """

    logger.debug("parsing array")
    arr = []
    while True:
        c = stream.read(1)

        if c in _GDB_MI_VALUE_START_CHARS:
            stream.seek(-1)
            val = _parse_val(stream)
            arr.append(val)
        elif c in _WHITESPACE:
            pass
        elif c == ",":
            pass
        elif c == "]":
            # Stop when this array has finished. Note
            # that elements of this array can be also be arrays.
            break

    logger.debug("parsed array:")
    logger.debug("%s", fmt_green(arr))
    return arr



from typing import List

from pygdbmi.gdbescapes import advance_past_string_with_gdb_escapes


__all__ = ["StringStream"]


class StringStream:
    """A simple class to hold text so that when passed
    between functions, the object is passed by reference
    and memory does not need to be repeatedly allocated for the string.

    This class was written here to avoid adding a dependency
    to the project.
    """

    def __init__(self, raw_text: str, debug: bool = False) -> None:
        self.raw_text = raw_text
        self.index = 0
        self.len = len(raw_text)

    def read(self, count: int) -> str:
        """Read count characters starting at self.index,
        and return those characters as a string
        """
        new_index = self.index + count
        if new_index > self.len:
            buf = self.raw_text[self.index :]  # return to the end, don't fail
        else:
            buf = self.raw_text[self.index : new_index]
        self.index = new_index

        return buf

    def seek(self, offset: int) -> None:
        """Advance the index of this StringStream by offset characters"""
        self.index = self.index + offset

    def advance_past_chars(self, chars: List[str]) -> str:
        """Advance the index past specific chars
        Args chars (list): list of characters to advance past

        Return substring that was advanced past
        """
        start_index = self.index
        while True:
            current_char = self.raw_text[self.index]
            self.index += 1
            if current_char in chars:
                break

            elif self.index == self.len:
                break

        return self.raw_text[start_index : self.index - 1]

    def advance_past_string_with_gdb_escapes(self) -> str:
        """Advance the index past a quoted string until the end quote is reached, and
        return the string (after unescaping it)

        Must be called only after encountering a quote character.
        """
        assert self.index > 0 and self.raw_text[self.index - 1] == '"', (
            "advance_past_string_with_gdb_escapes called not at the start of a string "
            f"(at index {self.index} of text {self.raw_text!r}, "
            f"remaining string {self.raw_text[self.index:]!r})"
        )

        unescaped_str, self.index = advance_past_string_with_gdb_escapes(
            self.raw_text, start=self.index
        )
        return unescaped_str



"""This module defines the `GdbController` class
which runs gdb as a subprocess and can write to it and read from it to get
structured output.
"""

import logging
import shutil
import subprocess
from typing import Dict, List, Optional, Union

from pygdbmi.constants import (
    DEFAULT_GDB_TIMEOUT_SEC,
    DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC,
)
from pygdbmi.IoManager import IoManager


__all__ = ["GdbController"]


DEFAULT_GDB_LAUNCH_COMMAND = ["gdb", "--nx", "--quiet", "--interpreter=mi3"]
logger = logging.getLogger(__name__)


class GdbController:
    def __init__(
        self,
        command: Optional[List[str]] = None,
        time_to_check_for_additional_output_sec: float = DEFAULT_TIME_TO_CHECK_FOR_ADDITIONAL_OUTPUT_SEC,
    ) -> None:
        """
        Run gdb as a subprocess. Send commands and receive structured output.
        Create new object, along with a gdb subprocess

        Args:
            command: Command to run in shell to spawn new gdb subprocess
            time_to_check_for_additional_output_sec: When parsing responses, wait this amout of time before exiting (exits before timeout is reached to save time). If <= 0, full timeout time is used.
        Returns:
            New GdbController object
        """

        if command is None:
            command = DEFAULT_GDB_LAUNCH_COMMAND

        if not any([("--interpreter=mi" in c) for c in command]):
            logger.warning(
                "Adding `--interpreter=mi3` (or similar) is recommended to get structured output. "
                + "See https://sourceware.org/gdb/onlinedocs/gdb/Mode-Options.html#Mode-Options."
            )
        self.abs_gdb_path = None  # abs path to gdb executable
        self.command: List[str] = command
        self.time_to_check_for_additional_output_sec = (
            time_to_check_for_additional_output_sec
        )
        self.gdb_process: Optional[subprocess.Popen] = None
        self._allow_overwrite_timeout_times = (
            self.time_to_check_for_additional_output_sec > 0
        )
        gdb_path = command[0]
        if not gdb_path:
            raise ValueError("a valid path to gdb must be specified")

        else:
            abs_gdb_path = shutil.which(gdb_path)
            if abs_gdb_path is None:
                raise ValueError(
                    'gdb executable could not be resolved from "%s"' % gdb_path
                )

            else:
                self.abs_gdb_path = abs_gdb_path

        self.spawn_new_gdb_subprocess()

    def spawn_new_gdb_subprocess(self) -> int:
        """Spawn a new gdb subprocess with the arguments supplied to the object
        during initialization. If gdb subprocess already exists, terminate it before
        spanwing a new one.
        Return int: gdb process id
        """
        if self.gdb_process:
            logger.debug(
                "Killing current gdb subprocess (pid %d)" % self.gdb_process.pid
            )
            self.exit()

        logger.debug(f'Launching gdb: {" ".join(self.command)}')

        # Use pipes to the standard streams
        self.gdb_process = subprocess.Popen(
            self.command,
            shell=False,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0,
        )

        assert self.gdb_process.stdin is not None
        assert self.gdb_process.stdout is not None
        self.io_manager = IoManager(
            self.gdb_process.stdin,
            self.gdb_process.stdout,
            self.gdb_process.stderr,
            self.time_to_check_for_additional_output_sec,
        )
        return self.gdb_process.pid

    def get_gdb_response(
        self,
        timeout_sec: float = DEFAULT_GDB_TIMEOUT_SEC,
        raise_error_on_timeout: bool = True,
    ) -> List[Dict]:
        """Get gdb response. See IoManager.get_gdb_response() for details"""
        return self.io_manager.get_gdb_response(timeout_sec, raise_error_on_timeout)

    def write(
        self,
        mi_cmd_to_write: Union[str, List[str]],
        timeout_sec: float = DEFAULT_GDB_TIMEOUT_SEC,
        raise_error_on_timeout: bool = True,
        read_response: bool = True,
    ) -> List[Dict]:
        """Write command to gdb. See IoManager.write() for details"""
        return self.io_manager.write(
            mi_cmd_to_write, timeout_sec, raise_error_on_timeout, read_response
        )

    def exit(self) -> None:
        """Terminate gdb process"""
        if self.gdb_process:
            self.gdb_process.terminate()
            self.gdb_process.wait()
            self.gdb_process.communicate()
        self.gdb_process = None
        return None
